/* 
*********************************************************************
relais_switch v 0.1 
*********************************************************************
toggles the relay board (8-fold serial) by conrad electronic;
by <m.ash@chello.at>

based on a modified version of winfried ritsch's comport external
<http://iem.kug.ac.at/ritsch/programmier/pd_externals/>
changes:
- removed polling, just send.
- dedicated to send 4 bytes messages including checksum needed by the card.

*********************************************************************
*/

#include "m_pd.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef NT
 #include <windows.h>
 #include <commctrl.h>
#else
 #include <sys/time.h>
 #include <fcntl.h> 
 #include <termios.h>               /* for TERMIO ioctl calls */
 #include <unistd.h>
 #define HANDLE int
 #define INVALID_HANDLE_VALUE -1
#endif

#include <string.h>
#include <errno.h>


static t_class *relais_switch_class;

typedef struct relais_switch
{
  t_object x_obj;
  HANDLE comhandle;
  char port_status;
  char print;
  
} t_relais_switch;




// -----------------------------------------------------
// OS-specific serial intitialisation and send functions
// -----------------------------------------------------



#ifdef NT
// -----------------------------------------------------
// WIN
// -----------------------------------------------------

static HANDLE comport_init (int com_nr)
{
	HANDLE com_port;
	DCB PortDCB;     
	COMMTIMEOUTS timeouts;
	#define COMPORT_MAX 8
	static char *sys_com_port[COMPORT_MAX] = {"COM1", "COM2", "COM3", "COM4","COM5", "COM6", "COM7", "COM8"};

  	if(com_nr < 0 || com_nr >= COMPORT_MAX) 
  	{
	 	post("relais_switch: comport %d not valid [0-8]",com_nr);
	 	return 0;
  	}
  
	com_port = CreateFile( sys_com_port[com_nr],  
		   GENERIC_READ | GENERIC_WRITE, 
		   0, 
		   NULL, 
		   OPEN_EXISTING,
		   0,
		   NULL);

	if(com_port == 0)
    {
      post("relais_switch: Error! Could not open serial device: %s",sys_com_port[com_nr]);
	  return 0;
    }

	if (!GetCommState(com_port, &PortDCB))
	{
		post("relais_switch: Error! Could not get state of serial device: %s", sys_com_port[com_nr]);
		CloseHandle(com_port);
		return 0;
	}
				
	PortDCB.DCBlength = sizeof (DCB); 
	GetCommState (com_port, &PortDCB);
	PortDCB.BaudRate = 19200;				// Current baud 
	PortDCB.fBinary = TRUE;					// Binary mode; no EOF check 
	PortDCB.fParity = TRUE;					// Enable parity checking 
	PortDCB.fOutxCtsFlow = FALSE;				// No CTS output flow control 
	PortDCB.fOutxDsrFlow = FALSE;				// No DSR output flow control 
//	PortDCB.com_porttrControl = DTR_CONTROL_DISABLE; 	// DTR flow control type 														
//	PortDCB.com_portsrSensitivity = FALSE;			// DSR sensitivity 
	PortDCB.fTXContinueOnXoff = FALSE;			// XOFF continues Tx 
	PortDCB.fOutX = FALSE;					// No XON/XOFF out flow control 
	PortDCB.fInX = FALSE;					// No XON/XOFF in flow control 
	PortDCB.fErrorChar = FALSE;				// Disable error replacement 
	PortDCB.fNull = FALSE;					// Disable null stripping 
	PortDCB.fRtsControl = RTS_CONTROL_DISABLE; 		// RTS flow control 
	PortDCB.fAbortOnError = FALSE;				// Do not abort reads/writes on error
	PortDCB.ByteSize = 8;					// Number of bits/byte, 4-8 
	PortDCB.Parity = NOPARITY;				// 0-4=no,odd,even,mark,space 
	PortDCB.StopBits = ONESTOPBIT;				// 0,1,2 = 1, 1.5, 2 

	if (!SetCommState(com_port, &PortDCB))
	{
		post("relais_switch: Error! Could not set state of serial device: %s", sys_com_port[com_nr]);
		return 0;
	}

	if (!GetCommTimeouts(com_port, &timeouts))
	{
	  post("relais_switch: Error! Could not get old timeouts for serial device: %s", sys_com_port[com_nr]);
	};

	//setting new timeouts for read to immidiatly return
	timeouts.ReadIntervalTimeout = MAXDWORD; 
	timeouts.ReadTotalTimeoutMultiplier = 0;
	timeouts.ReadTotalTimeoutConstant = 0;
	timeouts.WriteTotalTimeoutMultiplier = 0;
	timeouts.WriteTotalTimeoutConstant = 0;

	if (!SetCommTimeouts(com_port, &timeouts))
	{
  	  post("relais_switch: Error! Could not set timeouts for serial device: %s", sys_com_port[com_nr]);
	  return 0;
	};

	post("relais_switch: serial device %s opened.",sys_com_port[com_nr]);
	return com_port;	
}	
	
	
static void comport_close(t_relais_switch *x)
{
	CloseHandle(x->comhandle);
	post("relais_switch: serial port closed.");
}

void relais_switch_read(t_relais_switch *x)
{
	unsigned char chr;
	DWORD dwCommEvent;
  	DWORD dwRead;

	if (x->print) printf("relay answers: ");
 	do 
 	{
		if(ReadFile(x->comhandle, &chr, 1, &dwRead, NULL))
			if(dwRead > 0)
				if (x->print) printf("%d ",chr);
	    	else
	       		break;
	} 
	while (dwRead);
	if (x->print) printf("\n");
}

int relais_switch_send(t_relais_switch *x, char b1, char b2, char b3)
{
	char checksum;
	OVERLAPPED osWrite = {0};
	DWORD dwWritten = 0;
	DWORD dwBytesTransferred;   	
	
	checksum = b1 ^ b2 ^ b3;
	
	if (!WriteFile(x->comhandle,&b1, 1, &dwWritten, NULL)) return 0;
	if (!WriteFile(x->comhandle,&b2, 1, &dwWritten, NULL)) return 0;
	if (!WriteFile(x->comhandle,&b3, 1, &dwWritten, NULL)) return 0;
	if (!WriteFile(x->comhandle,&checksum, 1, &dwWritten, NULL)) return 0;
   	relais_switch_read(x);
    	// Sleep(10);
	return 1;
}





#else 

// -----------------------------------------------------
// LINUX/UNIX 
// -----------------------------------------------------


static HANDLE comport_init(int com_nr)
{
  	HANDLE fd;
  	struct termios *termiosinf = 0;

  	#define COMPORT_MAX 8
  	#define OPENPARAMS (O_RDWR|O_NDELAY|O_NOCTTY)
  	
  	static char *sys_com_port[8] = {"/dev/ttyS0", "/dev/ttyS1", "/dev/ttyS2", "/dev/ttyS3","/dev/ttyS4", "/dev/ttyS5", "/dev/ttyS6", "/dev/ttyS7"};

  	if(com_nr < 0 || com_nr >= COMPORT_MAX) 
  	{
		post("relais_switch: comport %d not valid [0-8]",com_nr);
	 	return INVALID_HANDLE_VALUE;
  	}

  	if ((fd = open(sys_com_port[com_nr], OPENPARAMS)) == INVALID_HANDLE_VALUE)
    	{
		post("relais_switch: Error! Could not open serial device: %s",sys_com_port[com_nr]);
		return INVALID_HANDLE_VALUE;
    	}

  	/* set no wait on any operation */
  	fcntl(fd, F_SETFL, FNDELAY);
  
  	/*   Save the Current Port Configuration  */
  	if(tcgetattr(fd, termiosinf) == -1)
	{
    		post("relais_switch: Error! could not get termios-structure of serial device %s\n",sys_com_port[com_nr]);
    		close(fd);
    		return INVALID_HANDLE_VALUE;
  	}
    
  	/* Setupt the new port configuration...NON-CANONICAL INPUT MODE
	 .. as defined in termios.h */
  	
  	/* enable input and ignore modem controls */
  	termiosinf->c_cflag |= (CREAD | CLOCAL);

  	/* always nocanonical, this means raw i/o no terminal */
  	termiosinf->c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);

  	/* no post processing */
	termiosinf->c_oflag &= ~OPOST;

  	/* defaults, see input */
	
	
  	//set_bits(x,8);      /* CS8 */
	termiosinf->c_cflag &= ~CSIZE;
	termiosinf->c_cflag |= CS8; 
	
  	//set_stopflag(x,0);  /* ~CSTOPB */
	
	
  	//set_ctsrts(x,0);  /* ~CRTSCTS;*/
  	//set_xonxoff(x,0); /* (IXON | IXOFF | IXANY) */


  	//set_baudrate(x,19200);
  	cfsetospeed(termiosinf, 19200);
	
  	if(tcsetattr(fd, TCSAFLUSH, termiosinf) != -1)
	{
		post("relais_switch: Opened serial line device  %s",sys_com_port[com_nr]);
	}
  	else 
	{
		post("relais_switch: Error! could not set params to ioctl of serial device %s",sys_com_port[com_nr]);
		close(fd);
		return INVALID_HANDLE_VALUE;
	}
	
  	return fd;
}


static void comport_close(t_relais_switch *x)
{
  	HANDLE fd = x->comhandle;

	if(fd != 0)
   	{
    	close(fd);
    	post("relais_switch: serial port closed.");
   	}
}

int relais_switch_send(t_relais_switch *x, char b1, char b2, char b3)
{
  if (!write(x->comhandle,&b1,1)) return 0;
  if (!write(x->comhandle,&b2,1)) return 0;
  if (!write(x->comhandle,&b3,1)) return 0;
  return 1;
}


#endif

// ------------------------------------------------------
// relais_switch main function for pd
// -----------------------------------------------------



int relais_init(t_relais_switch *x)
{
	if (!relais_switch_send(x,1,1,0)) return 0;
	sleep(10);
	if (!relais_switch_send(x,3,1,0)) return 0;
	return 1;
}


void relais_switch_toggle(t_relais_switch *x, t_floatarg f)
{
	char t = x->port_status ^ (char) pow(2,f-1);
	if (!relais_switch_send(x,3,1,t)) post("relais_switch: error sending");
	x->port_status = t;
}


void relais_switch_switchOn(t_relais_switch *x, t_floatarg f)
{
	char t = x->port_status | (char) pow(2,f-1);
	if (!relais_switch_send(x,3,1,t)) post("relais_switch: error sending");
	x->port_status = t;
}


void relais_switch_switchOff(t_relais_switch *x, t_floatarg f)
{
	char t = x->port_status & ~(char) pow(2,f-1);
	if (!relais_switch_send(x,3,1,t)) post("relais_switch: error sending");
	x->port_status = t;	
}


void relais_switch_allOff(t_relais_switch *x)
{
	if (!relais_switch_send(x,3,1,0)) post("relais_switch: error sending");
	x->port_status = 0;	
}


void relais_switch_command(t_relais_switch *x, t_floatarg f1, t_floatarg f2, t_floatarg f3)
{
	if (!relais_switch_send(x,(char)f1, (char)f2, (char)f3)) 
	post("relais_switch: error sending");
}


void relais_switch_list(t_relais_switch *x, t_symbol *s, int argc, t_atom *argv)
{
	if (argc==2)
	{
		if (atom_getfloatarg(1, argc, argv)>0) relais_switch_switchOn(x, atom_getfloatarg(0, argc, argv));
		else relais_switch_switchOff(x, atom_getfloatarg(0, argc, argv));
	}
	if (argc==3)
	{
		relais_switch_command(x, atom_getfloatarg(0, argc, argv), atom_getfloatarg(1, argc, argv), atom_getfloatarg(2, argc, argv));		
	}	
}


void relais_switch_float(t_relais_switch *x, t_floatarg f)
{
    if(f<=8 && f>0) relais_switch_toggle(x,f);
    else 
    	if (f==0) relais_switch_allOff(x);
    	else post("relais_switch: number is to high - currently only 8 ports available!");
}


void *relais_switch_new(t_floatarg f)
{
	t_relais_switch *x;
	int com_nr = f;
	
    x = (t_relais_switch *)pd_new(relais_switch_class);  
    x->comhandle = comport_init(com_nr);
    x->port_status = 0;
    x->print = 0;
    if (relais_init(x)) post("relais_switch: relais initialized");
    else post("relais_switch: error initializing");
    return x;
}


void relais_switch_close(t_relais_switch *x)
{
	relais_switch_allOff(x);
    comport_close(x);
}


void relais_switch_open(t_relais_switch *x, t_floatarg f)
{
	int com_nr = f;	
	
    comport_close(x);
    x->comhandle = comport_init(com_nr);
    relais_init(x);
}

void relais_switch_print(t_relais_switch *x, t_floatarg f)
{

    if (f>=0 && f<=1) x->print = (char) f;
}


void relais_switch_setup(void)
{
    relais_switch_class = class_new(gensym("relais_switch"), (t_newmethod)relais_switch_new, (t_method)relais_switch_close,
    	sizeof(t_relais_switch), CLASS_DEFAULT, A_DEFFLOAT, 0);  
    class_addfloat(relais_switch_class, relais_switch_float);
    class_addlist(relais_switch_class, relais_switch_list);
  	class_addmethod(relais_switch_class, (t_method)relais_switch_close, gensym("close"), 0);
  	class_addmethod(relais_switch_class, (t_method)relais_switch_open, gensym("open"),A_FLOAT, 0);    
  	class_addmethod(relais_switch_class, (t_method)relais_switch_switchOff, gensym("off"),A_FLOAT, 0);    
  	class_addmethod(relais_switch_class, (t_method)relais_switch_switchOn, gensym("on"),A_FLOAT, 0);    	
  	class_addmethod(relais_switch_class, (t_method)relais_switch_command, gensym("command"),A_FLOAT,A_FLOAT,A_FLOAT, 0);    	
  	class_addmethod(relais_switch_class, (t_method)relais_switch_print, gensym("print"),A_FLOAT, 0);    	
}

